
:- [txf_strategy].

:- import str_cat/3 from string.
:- import length/2, member/2, append/3 from basics.
:- dynamic proof_oblig/1.
:- dynamic unfold_prohibited/2.
:- dynamic no_unfold/1.
:- dynamic no_deduce/1.
:- dynamic proof_stack/2.
:- dynamic proved/2.
%---------------------------------------------
%  Setting the Constants/Flags

maxDepthConstant(8).   % maximum term depth to restrict infinite 
                       % call sequences during unfolding.  


fold_cond_max(2).      % Maximum length of conditions for 
                       % conditional folding.


equiv_cond_max(3).     % Maximum length of conditions for 
                       % conditional equivalence.

unfold_flag(suspend_expand).   % expand suspended atoms if they
                               % get instantiated.

unfold_flag(nonconsumer_freeze). % do not expand a clause body 
				 % if it has only open atoms.

%unfold_flag(suspend_freeze).  % atoms, once suspended can never
                               % be selected, even if instantiated.

%--------------------------------------------------------------------

% proof obligation : Pred1 <=> Pred2

prove(Pred1, Pred2) :-
	(proof_stack(Pred1, Pred2) ; proof_stack(Pred2, Pred1)), !, fail.

prove(Pred1, Pred2) :-
	( proved(Pred1, Pred2) ; proved(Pred2, Pred1) ), !.

prove(Pred1, Pred2) :-
	assert(proof_stack(Pred1, Pred2)),
	txf_prove(Pred1, Pred2),
	retract(proof_stack(Pred1, Pred2)),
	assert(proved(Pred1, Pred2)).

txf_prove(Pred1, Pred2) :-
	( axiom(Pred1, Pred2) 
		-> true
		;  transform(Pred1, Pred2), !,
		   txf_prove(Pred1, Pred2)
        ).


%--------------------------------------------------------------------------

% prove(Pred1, Pred2, Lo, Hi) is true if
%     1. Pred1 <=> Pred2
%     2. Lo \leq rank(Pred1) - rank(Pred2) \leq Hi


prove(Pred1, Pred2, Lo, Hi) :-
	assert(proof_stack(Pred1, Pred2)),
	txf_prove(Pred1, Pred2, Lo, Hi),
	retract(proof_stack(Pred1, Pred2)).


txf_prove(Pred1, Pred2, Lo, Hi) :-
	( axiom(Pred1, Pred2) 
		-> compute_proof_bounds(Pred1, Pred2, Lo, Hi)
		;  transform(Pred1, Pred2), !,
		   txf_prove(Pred1, Pred2, Lo, Hi)
        ).


%------------------------------------------------------------

transform(Pred1, Pred2) :- 
	choose_new_pred(Pred1, Pred2, Pred),
	find_clauses(Pred, CList), 
 	transform_clauses(Pred, CList),
	write_star(32),
	Pred1 = P/N, write_all(P, N),
	write_line(32),
	Pred2 = Q/M, write_all(Q, M),
	write_star(32).

/*
transform(Pred1, Pred2) :-
	  cond_axiom(Pred1, Pred2).
*/

transform_clauses(Pred, CList) :- 
	\+ no_unfold(Pred) -> algorithmic_transform(CList).
/*
transform_clauses(Pred, CList) :- 
	( \+ no_deduce(Pred) -> deductive_transform(CList)),
	( no_unfold(Pred) -> true 
			  ;  assert(no_unfold(Pred))
        ).

transform_clauses(Pred, _) :-
	( no_deduce(Pred) -> true
		          ;  assert(no_deduce(Pred))
        ), fail.
*/
algorithmic_transform(CList):-
	member(C, CList), select(C, Num), singleStepUnf(C, Num, CL),
	nl,writeln('UNFOLD'), write(C), write(' to '), writeln(CL),nl.

deductive_transform(CList) :- 
	cleanup_clauses(CList, CList1),
	( fold(CList1, _) ; 
	  cond_fold(CList1, _)
        ).

%-------------------------------------------------------------------
% A predicate is deemed new (and hence suitable for transformation) if
% it has been marked by user as a proof obligation predicate or it is
% a newly introduced predicate during transformation.

choose_new_pred(Pred1, Pred2, Pred) :-
	member(Pred, [Pred1, Pred2]), new_pred(Pred).


new_pred(P/N) :- proof_oblig(P/N).

new_pred(P/_) :- atom_chars(P, PList), PList = [n,e,w|_].

%---------------------------------------------------------------

