:- import conset/2, conget/2 from gensym.
:- import variant/2 from subsumes.
:- import append/3, member/2, length/2 from basics.
:- dynamic defn_pred/2.
:- dynamic oblig_arc/4.

%:- dynamic mark_new/1.

/*
goal_replace(Conj1/HV1, Conj2/HV2, Clause1, NewId) :-
	var_seteq(HV1, HV2),
	create_defn(Conj1, HV1, Pred1),
	create_defn(Conj2, HV2, Pred2),
	prove(Pred1, Pred2, D1, D2),!,
	update_proof_bounds(Conj1, Conj2, (D1, D2), (Delta1, Delta2)),
	Clause1 = clauseI(CId, Hd_Bd, (An1, An2)),
	gr_annot_chk(An1, Delta1, Hd_Bd),
	replace(Hd_Bd, Conj1, Conj2, Hd_Bdnew),
	plus_annot(An1, Delta1, AD1),
	plus_annot(An2, Delta2, AD2),
	retractall(clauseI(CId, _, _)), 
	gen_new_clauseNum(NewId),
	assert(clauseI(NewId, Hd_Bdnew, (AD1, AD2))).
*/
%---------------------------------------------------------------------

create_defn(Goal, HVars, Pred) :-
	exists_defn(Goal, HVars, Pred), !.

create_defn(Goal, HVars, Pred) :-
	( open_atom(Goal) -> atom_pred(Goal, P, N), Pred = P/N
			  ;  length(HVars, N), get_new_pred(P), 
                             Pred = P/N, Head =.. [P|HVars], 
                             define(P, N, [(Head :- Goal)])
        ).

exists_defn(Goal, HVars, Pred) :-
	clause0(C, (H :- B), _),
	functor(H, P, N), length(L, N), Hd =.. [P|L],
	findall(Cl, clause0(Cl, (Hd :- _), _), [C]),
	Head =.. [P|HVars], copy_term((H :- B), (Hnew :- Bnew)),
	exec_builtins(Goal, Goal1), exec_builtins(Bnew, B1),
        my_variant((Head :- Goal1), (Hnew :- B1)), Pred = P/N.

atom_pred(G, P, N) :-
	vanillaConvert(G, G1), functor(G1, P, N).

open_atom(G) :- vanillaConvert(G, G1), G1 =.. [_|L], var_list(L).

var_list([]).
var_list([V|Rest]) :- var(V), var_list(Rest).

%------------------------------------------------------
	
define(P, Arity, CList) :-
	CList = [_|_], 
	syntax_check(P, Arity, CList),
	%tell('NewDef'), writeln((P,Arity,CList)), told,
        define_clauses(CList, IdList),
	assert(defn_pred(P/Arity, IdList)).

syntax_check(_, _, []).
syntax_check(P, A, [C|Rest]) :-
	C = (H :- _), functor(H, P, A), syntax_check(P, A, Rest).


define_clauses([], []).
define_clauses([C|Rest], [N|Rest1]) :-
	gen_new_clauseNum(N), C = (H :- B), 
	assert_def(N, (H :- B), (1,1)),
	define_clauses(Rest, Rest1).

%---------------------------------------------------------------------

update_proof_bounds(A1, A2, (D1, D2), (D1, D2)) :-
	open_atom(A1), open_atom(A2), !.

update_proof_bounds(A1, A2, (D1, D2), (D1, D2)) :-
	\+ open_atom(A1), \+ open_atom(A2), !.

update_proof_bounds(A1, A2, (D1, D2), (Delta1, Delta2)) :-
	\+ open_atom(A1), open_atom(A2), !,
	minus_annot(D1, 1, Delta1), 
	minus_annot(D2, 1, Delta2). 

update_proof_bounds(A1, A2, (D1, D2), (Delta1, Delta2)) :-
	open_atom(A1), \+ open_atom(A2),
	plus_annot(D1, 1, Delta1), 
	plus_annot(D2, 1, Delta2). 

%-----------------------------------------------------------------
% the third condition for total correctness of Goal Replacement 
% given in the PPDP '99 paper. This condition allows us to prove 
% the total correctness of Goal Replacement by induction.

gr_annot_chk(Gamma, Delta, (_ :- Bd)) :-
	length(Bd, N), K is N - 1,       % replacing a single atom
	plus_annot(Gamma, Delta, GD),
	plus_annot(GD, K, GDK),
	greater_annot(GDK, 0).

%-----------------------------------------------------------------------

replace((Hd :- Bd), [], [], (Hd :- Bd)).

replace((Hd :- Bd), [A|Rest], [A1|Rest1], (Hd :- Bd1)) :-
	search_and_replace(Bd, A, A1, Bd_tmp),
	replace((Hd :- Bd_tmp), Rest, Rest1, (Hd :- Bd1)).

search_and_replace([A1|Rest1], Atom1, Atom2, [A2|Rest2]) :-
	( variant(A1, Atom1) 
                -> A1 = Atom1, A2 = Atom2, Rest2 = Rest1
                ;  A2 = A1,
                   search_and_replace(Rest1, Atom1, Atom2, Rest2)
        ).
	

%---------------------------------------------------------------
	
cond_axiom(Pred1, Pred2) :-
	( cleanup_and_axiom_chk(Pred1, Pred2)
		-> true
		;  find_clauses(Pred1, CL1),
		   find_clauses(Pred2, CL2),
		   gen_prove_conds(CL1, CL2, Pred1, Pred2),
	           cleanup_and_axiom_chk(Pred1, Pred2)
        ).


%  generate and prove conditions ; 
%  it is possible to exercise considerable choice in
%  finding the conditions to be proved; a syntactic equiv.
%  check is performed before cond_axiom succeeds.

gen_prove_conds([], _, _, _).

gen_prove_conds([C|Rest1], Cq, Pred1, Pred2) :- 
	compute_matching_clauses([C|Rest1], C, List1),
	%List1 = [C],
	compute_matching_clauses(Cq, C, List2),
	%hack_match(Cq, C, List2),
	%List2 = [D], 
	%clauseI(C,_,_), clauseI(D,_,_),
	( null_condition(List1, List2) 
		-> true
	        ; retrieve_pred(List1, New1/N),
		  retrieve_pred(List2, New2/N),
		  myassert(oblig_arc(Pred1,Pred2, New1/N, New2/N)),
		  conget(deduce_cnt, T), T1 is T + 1, 
		  conset(deduce_cnt, T1),
		  prove(New1/N, New2/N)
        ), remove_all(List1,[C|Rest1],Rest), 
	remove_all(List2, Cq, Cq_rest),
	gen_prove_conds(Rest, Cq_rest, Pred1, Pred2).
             

%------------------------------------------------------------


hack_match(List, C, L) :-
	get_clause(C, (H1 :- _), _),
	( (member(D, List), 
	   get_clause(D, (H2 :- _), _),
	   head_check(H1, H2)) ->  L = [D]
			       ;   L = []
        ).

%----------------------------------------------------------

compute_matching_clauses([], _C, []).

compute_matching_clauses([D|Rest], C, Match) :-
	get_clause(C, (H1 :- B1), _),	
	get_clause(D, (H2 :- B2), _),
	exec_builtins(B1, _), 
	exec_builtins(B2, _),
	( head_check(H1, H2) 
		-> M_i = [D]
		;  M_i = []
        ), append(M_i, Match1, Match),
	compute_matching_clauses(Rest, C, Match1).
	
retrieve_pred(CList, Newpred) :-
	create_defn(CList, Newpred),
	defn_fold(CList, Newpred).

head_check(H1, H2) :-
	H1 =.. [_|L1], H2 =.. [_|L2], my_variant(L2, L1).


defn_fold([], false/_) :- !.  % the pathological case : no-op

defn_fold([C], Pred) :-
	Pred = P/N, get_clause(C, (_ :- Bd), _),
	Bd = [Atom], length(Args, N), Atom =.. [P|Args], !.

defn_fold(CList, Pred) :- 
	fold(CList, Pred, _).

/*
	( ( proof_stack(Pred, _) ; proof_stack(_, Pred) )
	     -> conget(deduce_cnt, T), T1 is T+1, 
		conset(deduce_cnt, T1)
	     ;  true
        ).
*/

%------------------------------------------------------------

/*
null_condition(CList1, CList2) :-
	extract_bodies(CList1, pgm_i, BdList1),
	extract_bodies(CList2, pgm_i, BdList2),
	( (permute(BdList2, X), variant_list(X, BdList1)
          ) ; 
	  ( create_defn(CList1, New1/N),
	    create_defn(CList2, New2/N),
	    proof_stack(New1/N, New2/N),
	    defn_fold(CList1, New1/N),
	    defn_fold(CList2, New2/N)
          )
        ).
*/

/*
null_condition(CList1, CList2) :-
	  create_defn(CList1, New1/N),
	  create_defn(CList2, New2/N),
	  ( ( axiom(New1/N, New2/N)
              ; proof_stack(New1/N, New2/N)
            )  -> defn_fold(CList1, New1/N), 
	          defn_fold(CList2, New2/N)
	       ;  fail
          ).
*/


null_condition(CList1, CList2) :-
	extract_bodies(CList1, pgm_i, BdList1),
	extract_bodies(CList2, pgm_i, BdList2),
        ( (permute(BdList2, X), variant_list(X, BdList1))
		->  true
	        ;   ( create_defn(CList1, New1/N),
	              create_defn(CList2, New2/N),
	              ( ( proof_stack(New1/N,New2/N) ; 
			  axiom(New1/N,New2/N) 
                        )  -> collect_fold_stats(New1/N, New2/N),
			      defn_fold(CList1, New1/N), 
	                      defn_fold(CList2, New2/N)
	                   ;  fail
                      )
                    )
        ).       
        

collect_fold_stats(Pred1, Pred2) :-
	conget(deduce_cnt, T),
	( ( proof_stack(Pred1, _) ; proof_stack(_, Pred1) )
		-> T1 is T+1, conset(deduce_cnt, T1)
		;  true
        ), conget(deduce_cnt, T2),
	( ( proof_stack(Pred2, _) ; proof_stack(_, Pred2) )
		-> T3 is T2+1, conset(deduce_cnt, T3)
		;  true
        ).

%------------------------------------------------------------------

create_defn([], false/_) :- !.

create_defn(CList, Pred) :-
	exists_defn(CList, Pred), !.
	%assert(mark_new(Pred)).

create_defn(CList, P/N) :-
	extract_bodies(CList, pgm_i, [Bd/_]),
	Bd = [Atom], functor(Atom,P,N),
	Atom =.. [P|Args], var_list(Args), !.

create_defn(CList, Pred) :-
	get_new_pred(P),
	extract_bodies(CList, pgm_i, BdList),
	get_new_pred_arity(BdList, N),
	constr_def(P, BdList, DList),	
	define(P, N, DList), Pred = P/N.

get_new_pred_arity([_/HV], N) :- !, length(HV, N).
get_new_pred_arity([_/HV|Rest], N) :- 
	length(HV, N), get_new_pred_arity(Rest, N).


constr_def(_P, [], []).

constr_def(P, [Bd/HV | Rest], [(Hd :- Bd) | Rest1]) :-
	Hd =.. [P|HV], constr_def(P, Rest, Rest1).


exists_defn(CList, Pred) :-
	defn_pred(Pred, DList),
	length(CList,N), length(DList,N),
	extract_bodies(DList, pgm_0, BdList1),
	extract_bodies(CList, pgm_i, BdList2),
	permute(BdList1, X), variant_list(BdList2, X).

extract_bodies([], _, []).
extract_bodies([C|Rest], Status, [Bd1/HVars | Bdrest]) :-
	( Status == pgm_i 
	        -> get_clause(C, (Hd :- Bd), _)
		;  Status == pgm_0,
	           get_clause0(C, (Hd :- Bd), _)
        ), exec_builtins(Bd,Bd1), get_vars(Hd,HVars), 
	extract_bodies(Rest, Status, Bdrest).


%------------------------------------


variant_list([], []).
variant_list([H|T], [H1|T1]) :-
	my_variant(H, H1), variant_list(T, T1).

%-------------------------------------------

/*
impl_static(Pred1, Pred2) :-
	impl_static1(Pred1, Pred2) ; impl_static2(Pred1, Pred2).

impl_static1(Pred1, Pred2) :-
	abolish_all_tables, 
	proof_path(Pred1, X), 
	axiom(X, Pred2).

impl_static2(Pred1, Pred2) :-
	abolish_all_tables, 
	proof_path(X, Pred2), 
	axiom(Pred1, X).

:- table proof_path/2.

proof_path(X, X).
proof_path(X, Y) :- proved(X, Z), proof_path(Z, Y).
*/